// Jade MIF backend
// Copyright (c) 1997 Kathleen Marszalek
// Version 1.0 (beta 2b)
// Date: October 26, 1997

#include "MifDoc.h"
#include "ErrnoMessageArg.h"

#include <errno.h>

#ifdef SP_NAMESPACE
namespace SP_NAMESPACE {
#endif

struct MifMessages {
  // 5000
  static const MessageType2 cannotOpenOutputError;
  // 5001
  static const MessageType1 systemIdNotFilename;
  // 5002
  static const MessageType0 missingTableColumnFlowObject;
};

#ifdef SP_NAMESPACE
};
#endif

#ifdef DSSSL_NAMESPACE
namespace DSSSL_NAMESPACE {
#endif

// --------- static -----------------------------------------------------------

#define CONST_STRING( symbolName ) \
const String<char> MifDoc::s ## symbolName ## ( #symbolName, sizeof( #symbolName ) - 1 )

const String<char> MifDoc::sNONE;
const String<char> MifDoc::sSPACE( " ", sizeof( " " ) - 1 );
const String<char> MifDoc::sDefaultTblFormat
                    ( "Default Tbl Format", sizeof( "Default Tbl Format" ) - 1 );
const String<char> MifDoc::sDefaultPgfFormat
                    ( "Default Pgf Format", sizeof( "Default Pgf Format" ) - 1 );
const String<char> MifDoc::sTimesNewRoman
                    ( "Times New Roman", sizeof( "Times New Roman" ) - 1 );
const String<char> MifDoc::sPageNumXRefFormatName
                    ( "Page Number Format", sizeof( "Page Number Format" ) - 1 );
CONST_STRING( A );
CONST_STRING( Regular );
CONST_STRING( Bold );
CONST_STRING( Italic );
CONST_STRING( FNoUnderlining );
CONST_STRING( FSingle );
CONST_STRING( FDouble );
CONST_STRING( FNumeric );
CONST_STRING( FNormal );
CONST_STRING( FSuperscript );
CONST_STRING( FSubscript );
CONST_STRING( FAsTyped );
CONST_STRING( FSmallCaps );
CONST_STRING( FLowercase );
CONST_STRING( FUppercase );
CONST_STRING( LeftRight );
CONST_STRING( Left );
CONST_STRING( Right );
CONST_STRING( Center );
CONST_STRING( Fixed );
CONST_STRING( Proportional );
CONST_STRING( Top );
CONST_STRING( Middle );
CONST_STRING( Bottom );
CONST_STRING( NoLanguage );
CONST_STRING( USEnglish );
CONST_STRING( UKEnglish );
CONST_STRING( German );
CONST_STRING( SwissGerman );
CONST_STRING( French );
CONST_STRING( CanadianFrench );
CONST_STRING( Spanish );
CONST_STRING( Catalan );
CONST_STRING( Italian );
CONST_STRING( Portuguese );
CONST_STRING( Brazilian );
CONST_STRING( Danish );
CONST_STRING( Dutch );
CONST_STRING( Norwegian );
CONST_STRING( Nynorsk );
CONST_STRING( Finnish );
CONST_STRING( Swedish );
CONST_STRING( LeftMasterPage );
CONST_STRING( RightMasterPage );
CONST_STRING( OtherMasterPage );
CONST_STRING( BodyPage );
CONST_STRING( Decimal );
CONST_STRING( Tab );
CONST_STRING( HardSpace );
CONST_STRING( SoftHyphen );
CONST_STRING( HardHyphen );
CONST_STRING( HardReturn );
CONST_STRING( NumberSpace );
CONST_STRING( ThinSpace );
CONST_STRING( EnSpace );
CONST_STRING( EmSpace );
CONST_STRING( Header );
CONST_STRING( Footer );
CONST_STRING( Arabic );
CONST_STRING( UCRoman );
CONST_STRING( LCRoman );
CONST_STRING( UCAlpha );
CONST_STRING( LCAlpha );
CONST_STRING( FirstLeft );
CONST_STRING( FirstRight );
CONST_STRING( Anywhere );
CONST_STRING( ColumnTop );
CONST_STRING( PageTop );
CONST_STRING( LPageTop );
CONST_STRING( RPageTop );
CONST_STRING( First );
CONST_STRING( Inside );
CONST_STRING( Outside );
CONST_STRING( None );
CONST_STRING( RunIntoParagraph );
CONST_STRING( Inline );
CONST_STRING( Below );
CONST_STRING( Butt );
CONST_STRING( Round );
CONST_STRING( Square );
CONST_STRING( White );
CONST_STRING( Black );
CONST_STRING( Normal );
CONST_STRING( Straddle );
CONST_STRING( Cent );
CONST_STRING( Pound );
CONST_STRING( Yen );
CONST_STRING( EnDash );
CONST_STRING( EmDash );
CONST_STRING( Dagger );
CONST_STRING( DoubleDagger );
CONST_STRING( Bullet);

MifDoc *MifDoc::CurInstance = NULL;

MifDoc::T_ID MifDoc::Object::IDCnt = 0;

const char MifDoc::PotentialMarkerInfo::TypeChar = '\x1';
const char MifDoc::XRefInfo::TypeChar = '\x2';

MifDoc::T_indent MifOutputByteStream::INDENT = MifDoc::T_indent( 0xFFF );

bool MifDoc::Para::currentlyOpened = false;
unsigned MifDoc::ParaLine::setProperties = 0;
MifDoc::T_ID MifDoc::ParaLine::TextRectID;
MifDoc::T_ID MifDoc::ParaLine::ATbl;

// ----------------------------------------------------------------------------

int operator!=( const MifDoc::T_LTRB &o1, const MifDoc::T_LTRB &o2 ) {

    return o1.l != o2.l || o1.t != o2.t || o1.r != o2.r || o1.b != o2.b;
}

// --------- MifTmpOutputByteStream -------------------------------------------

void MifTmpOutputByteStream::commit( OutputByteStream &os, bool resolveCrossReferences ) {

    MifOutputByteStream outS( os );
    TmpOutputByteStream2::Iter iter( *this );
    const char *s;
    size_t n;

    MifDoc::ElementSet &elements = MifDoc::CurInstance->elements();
    MifDoc::PotentialMarkerInfo pmi;
    MifDoc::XRefInfo xri;
    char *byteDestPtr = NULL;
    size_t neededBytesNum = 0;
    enum { normal, waitingForTypeChar, inPotentialMarkerInfo, inXRefInfo } state = normal;
    while( iter.next( s, n ) ) {
        while( n > 0 ) {
            switch( state ) {

              case waitingForTypeChar:
                switch( *s ) {
                  case MifDoc::PotentialMarkerInfo::TypeChar:
                    neededBytesNum = sizeof( pmi );
                    byteDestPtr = (char*)&pmi;
                    state = inPotentialMarkerInfo;
                    break;
                  case MifDoc::XRefInfo::TypeChar:
                    neededBytesNum = sizeof( xri );
                    byteDestPtr = (char*)&xri;
                    state = inXRefInfo;
                    break;
                  default: assert( false );
                }
                n--; s++;
                break;

              case inPotentialMarkerInfo:
              case inXRefInfo: {
                int bytesCopiedNum = neededBytesNum > n ? n : neededBytesNum;
                memcpy( byteDestPtr, s, bytesCopiedNum );
                byteDestPtr += bytesCopiedNum; n -= bytesCopiedNum; s += bytesCopiedNum;
                neededBytesNum -= bytesCopiedNum;
                if( neededBytesNum == 0 ) {
                    if( state == inPotentialMarkerInfo ) {
                        if( elements.hasBeenReferenced( pmi.groveIndex, pmi.elementIndex ) ) {
                            outS.CurTagIndent = pmi.tagIndent;
                            MifDoc::Marker marker( MifDoc::Marker::makeText
                                                     ( pmi.groveIndex, pmi.elementIndex ) );
                            marker.out( outS );
                        }
                    } else { // inXRefInfo
                        outS.CurTagIndent = xri.tagIndent;
                        MifDoc::T_pathname targetFileName( "<c\\>" );
                        targetFileName
                         += MifDoc::T_pathname
                             ( MifDoc::CurInstance->bookComponents()
                                [elements.bookComponentIndex( xri.groveIndex, xri.elementIndex )]
                                  .FileName );
                        MifDoc::XRef xRef( MifDoc::sPageNumXRefFormatName,
                                           MifDoc::Marker::makeText
                                            ( xri.groveIndex, xri.elementIndex ),
                                           MifDoc::T_string( "000" ),
                                           targetFileName );
                       xRef.out( outS );
                    }
                    state = normal;
                }
                break; }

            case normal:
                if( resolveCrossReferences ) {
                    const char *escapeCharPtr
                     = (char*)memchr( s, MifDoc::escapeChar(), n );
                    if( escapeCharPtr != NULL ) {
                        outS.stream().sputn( s, escapeCharPtr - s);
                        n -= (escapeCharPtr - s); s = escapeCharPtr; n--; s++;
                        state = waitingForTypeChar;
                        break;
                    }
                }
                outS.stream().sputn( s, n );
                n = 0;
            }
        }
    }

    assert( neededBytesNum == 0 );
}

void MifTmpOutputByteStream::commit( String<char> &str ) {

    TmpOutputByteStream2::Iter iter( *this );
    const char *s;
    size_t n;

    while( iter.next( s, n ) ) {
        str.append( s, n );
    }
}

// ----------------------------------------------------------------------------

#define CHECK_PROPERTY( propertyName ) \
    \
    if( setProperties & f ## propertyName ) \
        os << '\n' << MifOutputByteStream::INDENT \
           << "<" #propertyName " " << propertyName << ">"; 

#define CHECK_VECTOR_OF_PROPERTIES( propertyName ) \
    \
    if( setProperties & f ## propertyName ## s) \
        for( size_t i = 0; i < propertyName ## s.size(); i++ ) { \
            os << '\n' << MifOutputByteStream::INDENT \
               << "<" #propertyName " " << propertyName ## s[i] << ">"; \
        }

// --------- MifDoc::TagStream ------------------------------------------------

MifDoc::TagStream::TagStream( MifDoc::T_indent osIndent )
 : Format(), Content( new MifTmpOutputByteStream( osIndent ) ), PgfTagUsed( false ),
   InitialPgfTag( sDefaultPgfFormat ), TagStreamClass( tsTagStream ) {
}

MifDoc::TagStream::~TagStream() {

    if( Content )
        delete Content;
}

// --------- MifDoc::TextFlow -------------------------------------------------

MifDoc::TextFlow::TextFlow( const TextRect &textRect, bool body, ParagraphFormat *format, 
                            T_tagstring pgfTag )
 : TagStream(), TextRectID( textRect.ID ), setProperties( 0 ), TextRectIDUsed( false ),
   TextRect_( &textRect ) {

    TagStreamClass = tsTextFlow;
    InitialPgfTag = pgfTag;

    if( format )
        Format = *format;

    if( body ) {
        setTFTag( MifDoc::sA );
        setTFAutoConnect( true );
    }
}

void MifDoc::TextFlow::out( MifOutputByteStream &os, bool resolveCrossReferences ) {

    os << '\n' << MifOutputByteStream::INDENT << "<TextFlow ";
    os.indent();
    CHECK_PROPERTY( TFTag );
    CHECK_PROPERTY( TFAutoConnect );
    content().commit( os.stream(), resolveCrossReferences );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Ruling ---------------------------------------------------

const String<char> &MifDoc::Ruling::key( Ruling &r ) {

    if( r.Key.size() < 1 ) {
        char keyStr[ 48 ];
        sprintf( keyStr, "Ruling-%li-%li-%li-%li",
                 long( r.RulingPenWidth ), long( r.RulingGap ), r.RulingPen, r.RulingLines );
        r.Key.assign( keyStr, strlen( keyStr ) );
     }

     return r.Key;       
}

void MifDoc::Ruling::out( MifOutputByteStream &os ) const {

    os << '\n' << MifOutputByteStream::INDENT << "<Ruling ";
    os.indent();
    CHECK_PROPERTY( RulingTag );
    CHECK_PROPERTY( RulingPenWidth );
    CHECK_PROPERTY( RulingGap );
    CHECK_PROPERTY( RulingPen );
    CHECK_PROPERTY( RulingLines );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::RulingCatalog --------------------------------------------

void MifDoc::RulingCatalog::out( MifOutputByteStream &os ) {

    if( Rulings.count() > 0 ) {
        os << '\n' << MifOutputByteStream::INDENT << "<RulingCatalog ";
        os.indent();
    
        PointerTableIter<Ruling*,String<char>,StringHash,Ruling> iter( Rulings );
        const Ruling *ruling;
        while( ( ruling = iter.next() ) != NULL )
            ruling->out( os );
    
        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }
}

// --------- MifDoc::Cell -----------------------------------------------------

void MifDoc::Cell::out( MifOutputByteStream &os, bool resolveCrossReferences ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Cell ";
    os.indent();

    CHECK_PROPERTY( CellLRuling );
    CHECK_PROPERTY( CellBRuling );
    CHECK_PROPERTY( CellRRuling );
    CHECK_PROPERTY( CellTRuling );
    CHECK_PROPERTY( CellColumns );
    CHECK_PROPERTY( CellRows );

    os << '\n' << MifOutputByteStream::INDENT << "<CellContent ";
    os.indent();
    content().commit( os.stream(), resolveCrossReferences );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Row ------------------------------------------------------

void MifDoc::Row::out( MifOutputByteStream &os, bool resolveCrossReferences ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Row ";
    os.indent();

    for( size_t i = 0; i < Cells.size(); i++ ) Cells[i].out( os, resolveCrossReferences );

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::TblColumn ------------------------------------------------

void MifDoc::TblColumn::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<TblFormat ";
    os.indent();

    CHECK_PROPERTY( TblColumnNum );
    CHECK_PROPERTY( TblColumnWidth );

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::TblFormat ------------------------------------------------

void MifDoc::TblFormat::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<TblFormat ";
    os.indent();

    CHECK_PROPERTY( TblTag );
    CHECK_PROPERTY( TblLIndent );
    CHECK_PROPERTY( TblRIndent );
    CHECK_PROPERTY( TblSpBefore );
    CHECK_PROPERTY( TblSpAfter );
    CHECK_PROPERTY( TblAlignment );
    CHECK_PROPERTY( TblCellMargins );
    CHECK_PROPERTY( TblTitlePlacement );
    CHECK_PROPERTY( TblWidth );
    for( size_t i = 0; i < TblColumns.size(); i++ ) TblColumns[i].out( os );

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Tbl ------------------------------------------------------

void MifDoc::Tbl::out( MifOutputByteStream &os, bool resolveCrossReferences ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Tbl ";
    os.indent();
    CHECK_PROPERTY( TblID );
    CHECK_PROPERTY( TblTag );

    if( setProperties & fTblFormat )
        TblFormat.out( os );

    CHECK_PROPERTY( TblNumColumns );
    CHECK_VECTOR_OF_PROPERTIES( TblColumnWidth );

    size_t i;

    if( TblH.size() ) {
        os << '\n' << MifOutputByteStream::INDENT << "<TblH ";
        os.indent();
        for( i = 0; i < TblH.size(); i++ ) TblH[i].out( os, resolveCrossReferences );
        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }

    if( TblBody.size() ) {
        os << '\n' << MifOutputByteStream::INDENT << "<TblBody ";
        os.indent();
        for( i = 0; i < TblBody.size(); i++ ) TblBody[i].out( os, resolveCrossReferences );
        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }

    if( TblF.size() ) {
        os << '\n' << MifOutputByteStream::INDENT << "<TblF ";
        os.indent();
        for( i = 0; i < TblF.size(); i++ ) TblF[i].out( os, resolveCrossReferences );
        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }
    
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Document -------------------------------------------------

void MifDoc::Document::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Document ";
    os.indent();
    CHECK_PROPERTY( DMargins );
    CHECK_PROPERTY( DColumns );
    CHECK_PROPERTY( DColumnGap );
    CHECK_PROPERTY( DPageSize );
    CHECK_PROPERTY( DStartPage );
    CHECK_PROPERTY( DPageNumStyle );
    CHECK_PROPERTY( DTwoSides );
    CHECK_PROPERTY( DParity );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::XRefFormat -----------------------------------------------

void MifDoc::XRefFormat::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<XRefFormat ";
    os.indent();
    CHECK_PROPERTY( XRefName );
    CHECK_PROPERTY( XRefDef );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::XRef -----------------------------------------------------

void MifDoc::XRef::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<XRef ";
    os.indent();
    CHECK_PROPERTY( XRefName );
    CHECK_PROPERTY( XRefSrcText );
    CHECK_PROPERTY( XRefSrcFile );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
    if( setProperties & fXRefText )
        os << '\n' << MifOutputByteStream::INDENT
           << "<String " << XRefText << ">"; 
    os << '\n' << MifOutputByteStream::INDENT << "<XRefEnd>";
}

// --------- MifDoc::Marker ---------------------------------------------------

void MifDoc::Marker::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Marker ";
    os.indent();
    CHECK_PROPERTY( MType );
    CHECK_PROPERTY( MText );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

MifDoc::T_string MifDoc::Marker::makeText
 ( unsigned long groveIndex, unsigned long elementIndex ) {

    char buf[32]; sprintf( buf, "NODE %lu.%lu", groveIndex, elementIndex );
    return T_string( buf );
}

// --------- MifDoc::Object ---------------------------------------------------

void MifDoc::Object::outObjectProperties( MifOutputByteStream &os ) {

    CHECK_PROPERTY( ID );
    CHECK_PROPERTY( Pen );
    CHECK_PROPERTY( Fill );
    CHECK_PROPERTY( PenWidth );
}

// --------- MifDoc::PolyLine -------------------------------------------------

void MifDoc::PolyLine::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<PolyLine ";
    os.indent();
    outObjectProperties( os );
    CHECK_PROPERTY( HeadCap );
    CHECK_PROPERTY( TailCap );
    for( size_t i = 0; i < Points.size(); i++ )
        os << '\n' << MifOutputByteStream::INDENT << "<Point " << Points[i] << ">";
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::ImportObject ---------------------------------------------

void MifDoc::ImportObject::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<ImportObject ";
    os.indent();
    outObjectProperties( os );
    CHECK_PROPERTY( ImportObFileDI );
    CHECK_PROPERTY( BitMapDpi );
    CHECK_PROPERTY( ShapeRect );
    CHECK_PROPERTY( ImportObFixedSize );
    CHECK_PROPERTY( NativeOrigin );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Frame ----------------------------------------------------

void MifDoc::Frame::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Frame ";
    os.indent();
    outObjectProperties( os );
    CHECK_PROPERTY( ShapeRect );
    CHECK_PROPERTY( FrameType );
    CHECK_PROPERTY( NSOffset );
    CHECK_PROPERTY( BLOffset );
    CHECK_PROPERTY( AnchorAlign );
    for( size_t i = 0; i < Objects.size(); i++ ) Objects[i]->out( os );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::TextRect -------------------------------------------------

void MifDoc::TextRect::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<TextRect ";
    os.indent();
    outObjectProperties( os );
    CHECK_PROPERTY( ShapeRect );
    CHECK_PROPERTY( TRNumColumns );
    CHECK_PROPERTY( TRColumnGap );
    CHECK_PROPERTY( TRColumnBalance );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Page -----------------------------------------------------

void MifDoc::Page::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<Page ";
    os.indent();
    CHECK_PROPERTY( PageType );
    CHECK_PROPERTY( PageTag );
    CHECK_PROPERTY( PageBackground );
    for( size_t i = 0; i < TextRects.size(); i++ )
        TextRects[i].out( os );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}                                      

// --------- MifDoc::TabStop --------------------------------------------------

void MifDoc::TabStop::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<TabStop ";
    os.indent();
    CHECK_PROPERTY( TSX );
    CHECK_PROPERTY( TSType );
    CHECK_PROPERTY( TSLeaderStr );
    CHECK_PROPERTY( TSDecimalChar );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::Para -----------------------------------------------------

void MifDoc::Para::out( MifOutputByteStream &os ) {

    outProlog( os );
    MifDoc::ParaLine::outProlog( os );
    content().commit( os.stream() );
    MifDoc::ParaLine::outEpilog( os );
    outEpilog( os );
}

void MifDoc::Para::outSimpleProlog( MifOutputByteStream &os ) {

    currentlyOpened = true;
    os << '\n' << MifOutputByteStream::INDENT << "<Para ";
    os.indent();
}

void MifDoc::Para::outProlog( MifOutputByteStream &os ) {

    currentlyOpened = true;
    outSimpleProlog( os );
    CHECK_PROPERTY( PgfTag );
    if( setProperties & fParagraphFormat )
        format().out( os );
}

void MifDoc::Para::outEpilog( MifOutputByteStream &os ) {

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
    currentlyOpened = false;
}

// --------- MifDoc::ParaLine -------------------------------------------------

void MifDoc::ParaLine::outProlog( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<ParaLine ";
    os.indent();
    
    MifDoc::TextFlow *tf = MifDoc::CurInstance->curTextFlow();
    if( tf != NULL && !tf->TextRectIDUsed ) {
        setTextRectID( tf->TextRectID );
        CHECK_PROPERTY( TextRectID );
        setProperties &= ~fTextRectID;
        tf->TextRectIDUsed = true;
    }

    MifDoc::Tbl *tbl = MifDoc::CurInstance->curTbl( false );
    if( tbl != NULL && !tbl->TblIDUsed ) {
        setATbl( tbl->TblID );
        CHECK_PROPERTY( ATbl );
        setProperties &= ~fATbl;
        tbl->TblIDUsed = true;
    }
}

void MifDoc::ParaLine::outEpilog( MifOutputByteStream &os ) {

    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}


// --------- MifDoc::PgfCatalog -----------------------------------------------

void MifDoc::PgfCatalog::out( MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<PgfCatalog ";
    os.indent();
    for( size_t i = 0; i < ParaFormats.size(); i++ ) ParaFormats[i].out( os, false );
    os.undent();
    os << '\n' << MifOutputByteStream::INDENT << ">";
}

// --------- MifDoc::TblCatalog -----------------------------------------------

void MifDoc::TblCatalog::out( MifOutputByteStream &os ) {

    if( TblFormats.size() > 0 ) {
        os << '\n' << MifOutputByteStream::INDENT << "<TblCatalog ";
        os.indent();
        for( size_t i = 0; i < TblFormats.size(); i++ ) TblFormats[i].out( os );
        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }
}

// ----------------------------------------------------------------------------

#define COMPARE_PROPERTY( propertyName ) \
\
    if( propertyName != f. ## propertyName ) \
        differingProperties |= f ## propertyName;        

#define COND_SET_PROPERTY( propertyName ) \
\
    if( properties & f ## propertyName ) \
        propertyName =  f. ## propertyName;        

#define PROPERTY_OUT( propertyName ) \
\
    if( properties & f ## propertyName ) \
        os << '\n' << MifOutputByteStream::INDENT \
           << "<" #propertyName " " << propertyName << ">"; 

// --------- MifDoc::FontFormat -----------------------------------------------

unsigned MifDoc::FontFormat::compare( FontFormat &f ) {

    unsigned differingProperties = 0;

    COMPARE_PROPERTY( FFamily );
    COMPARE_PROPERTY( FAngle );
    COMPARE_PROPERTY( FWeight );
    COMPARE_PROPERTY( FVar );
    COMPARE_PROPERTY( FSize );
    COMPARE_PROPERTY( FColor );
    COMPARE_PROPERTY( FUnderlining );
    COMPARE_PROPERTY( FOverline );
    COMPARE_PROPERTY( FStrike );
    COMPARE_PROPERTY( FPosition );
    COMPARE_PROPERTY( FPairKern );
    COMPARE_PROPERTY( FCase );
    COMPARE_PROPERTY( FDX );
    COMPARE_PROPERTY( FDY );
    COMPARE_PROPERTY( FDW );

    return differingProperties;
}

void MifDoc::FontFormat::setFrom( FontFormat &f, unsigned properties ) {

    COND_SET_PROPERTY( FFamily );
    COND_SET_PROPERTY( FAngle );
    COND_SET_PROPERTY( FWeight );
    COND_SET_PROPERTY( FVar );
    COND_SET_PROPERTY( FSize );
    COND_SET_PROPERTY( FColor );
    COND_SET_PROPERTY( FUnderlining );
    COND_SET_PROPERTY( FOverline );
    COND_SET_PROPERTY( FStrike );
    COND_SET_PROPERTY( FPosition );
    COND_SET_PROPERTY( FPairKern );
    COND_SET_PROPERTY( FCase );
    COND_SET_PROPERTY( FDX );
    COND_SET_PROPERTY( FDY );
    COND_SET_PROPERTY( FDW );
}

void MifDoc::FontFormat::out( MifOutputByteStream &os,
                              unsigned properties,
                              FontStatement fontStatement ) {

    const char *statementName = (fontStatement == stPgfFont) ? "PgfFont" : "Font";

    if( properties ) {

        os << '\n' << MifOutputByteStream::INDENT << "<" << statementName << " ";
        os.indent();

        PROPERTY_OUT( FFamily );
        PROPERTY_OUT( FAngle );
        PROPERTY_OUT( FWeight );
        PROPERTY_OUT( FVar );
        PROPERTY_OUT( FSize );
        PROPERTY_OUT( FColor );
        PROPERTY_OUT( FUnderlining );
        PROPERTY_OUT( FOverline );
        PROPERTY_OUT( FStrike );
        PROPERTY_OUT( FPosition );
        PROPERTY_OUT( FPairKern );
        PROPERTY_OUT( FCase );
        PROPERTY_OUT( FDX );
        PROPERTY_OUT( FDY );
        PROPERTY_OUT( FDW );

        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }
}

void MifDoc::FontFormat::updateFrom( FontFormat &f ) {

    unsigned differingProperties = FontFormat::compare( f );
    setFrom( f, differingProperties );
    setProperties = differingProperties;
}

// --------- MifDoc::ParagraphFormat ------------------------------------------

unsigned MifDoc::ParagraphFormat::compare( ParagraphFormat &f ) {

    unsigned differingProperties = 0;

    COMPARE_PROPERTY( PgfTag );
    COMPARE_PROPERTY( PgfLanguage );
    COMPARE_PROPERTY( PgfFIndent );
    COMPARE_PROPERTY( PgfLIndent );
    COMPARE_PROPERTY( PgfRIndent );
    COMPARE_PROPERTY( PgfAlignment );
    COMPARE_PROPERTY( PgfSpBefore );
    COMPARE_PROPERTY( PgfSpAfter );
    COMPARE_PROPERTY( PgfLineSpacing );
    COMPARE_PROPERTY( PgfLeading );
    COMPARE_PROPERTY( PgfNumTabs );
    COMPARE_PROPERTY( PgfPlacement );
    COMPARE_PROPERTY( PgfPlacementStyle );
    COMPARE_PROPERTY( PgfWithPrev );
    COMPARE_PROPERTY( PgfWithNext );
    COMPARE_PROPERTY( PgfBlockSize );
    COMPARE_PROPERTY( PgfAutoNum );
    COMPARE_PROPERTY( PgfNumFormat );
    COMPARE_PROPERTY( PgfNumberFont );
    COMPARE_PROPERTY( PgfHyphenate );
    COMPARE_PROPERTY( HyphenMaxLines );
    COMPARE_PROPERTY( HyphenMinPrefix );
    COMPARE_PROPERTY( HyphenMinSuffix );
    COMPARE_PROPERTY( HyphenMinWord );
    COMPARE_PROPERTY( PgfLetterSpace );
    COMPARE_PROPERTY( PgfCellAlignment );
    COMPARE_PROPERTY( PgfCellMargins );
    COMPARE_PROPERTY( PgfCellLMarginFixed );
    COMPARE_PROPERTY( PgfCellTMarginFixed );
    COMPARE_PROPERTY( PgfCellRMarginFixed );
    COMPARE_PROPERTY( PgfCellBMarginFixed );

    return differingProperties;
}

void MifDoc::ParagraphFormat::setFrom( ParagraphFormat &f,
                                       unsigned properties,
                                       unsigned fontProperties ) {
    COND_SET_PROPERTY( PgfTag );
    COND_SET_PROPERTY( PgfLanguage );
    COND_SET_PROPERTY( PgfFIndent );
    COND_SET_PROPERTY( PgfLIndent );
    COND_SET_PROPERTY( PgfRIndent );
    COND_SET_PROPERTY( PgfAlignment );
    COND_SET_PROPERTY( PgfSpBefore );
    COND_SET_PROPERTY( PgfSpAfter );
    COND_SET_PROPERTY( PgfLineSpacing );
    COND_SET_PROPERTY( PgfLeading );
    COND_SET_PROPERTY( PgfNumTabs );
    COND_SET_PROPERTY( PgfPlacement );
    COND_SET_PROPERTY( PgfPlacementStyle );
    COND_SET_PROPERTY( PgfWithPrev );
    COND_SET_PROPERTY( PgfWithNext );
    COND_SET_PROPERTY( PgfBlockSize );
    COND_SET_PROPERTY( PgfAutoNum );
    COND_SET_PROPERTY( PgfNumFormat );
    COND_SET_PROPERTY( PgfNumberFont );
    COND_SET_PROPERTY( PgfHyphenate );
    COND_SET_PROPERTY( HyphenMaxLines );
    COND_SET_PROPERTY( HyphenMinPrefix );
    COND_SET_PROPERTY( HyphenMinSuffix );
    COND_SET_PROPERTY( HyphenMinWord );
    COND_SET_PROPERTY( PgfLetterSpace );
    COND_SET_PROPERTY( PgfCellAlignment );
    COND_SET_PROPERTY( PgfCellMargins );
    COND_SET_PROPERTY( PgfCellLMarginFixed );
    COND_SET_PROPERTY( PgfCellTMarginFixed );
    COND_SET_PROPERTY( PgfCellRMarginFixed );
    COND_SET_PROPERTY( PgfCellBMarginFixed );

    FontFormat::setFrom( f, fontProperties );
}

void MifDoc::ParagraphFormat::out( MifOutputByteStream &os,
                                   unsigned properties,
                                   unsigned fontProperties,
                                   bool excludeCellProperties ) {

    bool outPgfTag = false;
    T_tagstring pgfTag;

    if( !( properties & fPgfTag ) ) {
        MifDoc::TagStream &ts = MifDoc::CurInstance->curTagStream();
        if( !ts.PgfTagUsed ) {
            pgfTag = ts.InitialPgfTag;
            outPgfTag = true;
            ts.PgfTagUsed = true;
        }
    }

    if( properties != 0 | fontProperties != 0 | outPgfTag ) {

        os << '\n' << MifOutputByteStream::INDENT << "<Pgf ";
        os.indent();

        if( outPgfTag )
          os << '\n' << MifOutputByteStream::INDENT << "<PgfTag " << pgfTag << ">"; 
        else 
          PROPERTY_OUT( PgfTag );

        FontFormat::out( os, fontProperties, FontFormat::stPgfFont );

        PROPERTY_OUT( PgfLanguage );
        PROPERTY_OUT( PgfFIndent );
        PROPERTY_OUT( PgfLIndent );
        PROPERTY_OUT( PgfRIndent );
        PROPERTY_OUT( PgfAlignment );
        PROPERTY_OUT( PgfSpBefore );
        PROPERTY_OUT( PgfSpAfter );
        PROPERTY_OUT( PgfLineSpacing );
        PROPERTY_OUT( PgfLeading );

        PROPERTY_OUT( PgfNumTabs );
        if( properties & fTabStops )
            for( size_t i = 0; i < TabStops.size(); i++ ) TabStops[i].out( os );

        PROPERTY_OUT( PgfPlacement );
        PROPERTY_OUT( PgfPlacementStyle );
        PROPERTY_OUT( PgfWithPrev );
        PROPERTY_OUT( PgfWithNext );
        PROPERTY_OUT( PgfBlockSize );
        PROPERTY_OUT( PgfAutoNum );
        PROPERTY_OUT( PgfNumFormat );
        PROPERTY_OUT( PgfNumberFont );
        PROPERTY_OUT( PgfHyphenate );
        PROPERTY_OUT( HyphenMaxLines );
        PROPERTY_OUT( HyphenMinPrefix );
        PROPERTY_OUT( HyphenMinSuffix );
        PROPERTY_OUT( HyphenMinWord );
        PROPERTY_OUT( PgfLetterSpace );

        if( MifDoc::CurInstance->curCell() != NULL || !excludeCellProperties ) {

            PROPERTY_OUT( PgfCellAlignment );
            PROPERTY_OUT( PgfCellMargins );
            PROPERTY_OUT( PgfCellLMarginFixed );
            PROPERTY_OUT( PgfCellTMarginFixed );
            PROPERTY_OUT( PgfCellRMarginFixed );
            PROPERTY_OUT( PgfCellBMarginFixed );
        }

        os.undent();
        os << '\n' << MifOutputByteStream::INDENT << ">";
    }
}

void MifDoc::ParagraphFormat::updateFrom( ParagraphFormat &f ) {

    FontFormat::updateFrom( f );

    unsigned differingProperties = compare( f );
    setFrom( f, differingProperties, 0 );
    setProperties = differingProperties;
}

// --------- MifDoc::BookComponent --------------------------------------------

MifDoc::BookComponent::BookComponent( MifTmpOutputByteStream *epilogOs )
  : epilogOs( epilogOs ), pageNumXRefFormatGenerated( false ) {
}  

void MifDoc::BookComponent::commit
( const String<CmdLineApp::AppChar> &dirName, bool resolveCrossReferences ) {

    assert( FileName.size() > 0 );
    String<CmdLineApp::AppChar> fileLoc;
    fileLoc += dirName; fileLoc += FileName; fileLoc += 0;
    FileOutputByteStream outFile;
    if( outFile.open( fileLoc.data() ) ) {
    
        MifOutputByteStream os( outFile );

        os << "<MIFFile 5.0>";

        PgfCatalog.out( os );
        RulingCatalog.out( os );

        size_t i;
        if( AFrames.size() ) {
            os << "\n<AFrames ";
            os.indent();
            for( i = 0; i < AFrames.size(); i++ ) AFrames[i].out( os );
            os.undent();
            os << '\n' << ">";
        }

        if( XRefFormats.size() ) {
            os << "\n<XRefFormats ";
            os.indent();
            for( i = 0; i < XRefFormats.size(); i++ ) XRefFormats[i].out( os );
            os.undent();
            os << '\n' << ">";
        }

        TblCatalog.out( os );
        Document.out( os );
    
        if( Tbls.size() ) {
            os << "\n<Tbls ";
            os.indent();
            for( i = 0; i < Tbls.size(); i++ ) Tbls[i].out( os, resolveCrossReferences );
            os.undent();
            os << '\n' << ">";
        }
    
        for( i = 0; i < Pages.size(); i++ ) Pages[i].out( os );
        for( i = 0; i < TextFlows.size(); i++ ) TextFlows[i].out( os, resolveCrossReferences );
    
        if( epilogOs )
            epilogOs->commit( os.stream(), resolveCrossReferences );    

    } else {
        MifDoc::curInstance().App->message
         ( MifMessages::cannotOpenOutputError,
           StringMessageArg( MifDoc::curInstance().App->convertInput( fileLoc.data() ) ),
           ErrnoMessageArg( errno ) );
        return;
    }
}

// --------- MifDoc -----------------------------------------------------------

MifDoc::MifDoc( const String<CmdLineApp::AppChar> &fileLoc, CmdLineApp *app )
 : App( app ), RootOutputFileLoc( fileLoc ), CurTblNum( 0 ), CurTextFlow( NULL ),
   CurCell( NULL ), CurPara( NULL ), NextID( 0 ) {

    CurInstance = this;
    enterBookComponent();
}

MifDoc::~MifDoc() {

    commit();
}

void MifDoc::commit() {

    String<CmdLineApp::AppChar> outDir = rootOutputFileLoc();
    String<CmdLineApp::AppChar> outFileName;

    { int i;
    for( i = outDir.size() - 1; i >= 0; i-- )
        if( outDir[i] == '/' || outDir[i] == '\\' )
            break;
    if( outDir.size() - i > 0 )
        outFileName.assign( &outDir.data()[i+1], outDir.size() - i );
    outDir.resize( i + 1 ); }
  
    if( BookComponents.size() > 1 ) {
        String<CmdLineApp::AppChar> fileName;
        char idxStr[32];
        for( size_t i = 0; i < BookComponents.size(); i++ ) {
            fileName.resize( 0 );          
            sprintf( idxStr, "%u", i+1 );
            for( const char *c = idxStr; *c; c++)
                fileName += CmdLineApp::AppChar(*c);
            fileName += '.'; fileName += 'm'; fileName += 'i'; fileName += 'f';
            BookComponents[i].FileName = fileName;  
        }

        String<CmdLineApp::AppChar> bookFileLoc( rootOutputFileLoc() );
        bookFileLoc += 0;
        FileOutputByteStream bookFile;
        if( bookFile.open( bookFileLoc.data() ) ) {
    
            MifOutputByteStream os( bookFile );
            os << "<Book 5.0>";
            for( size_t i = 0; i < BookComponents.size(); i++ )
                os << "\n<BookComponent"
                   << "\n  <FileName `<c\\>" << BookComponents[i].FileName << "\'>"
                   << "\n>";
        } else {

            MifDoc::curInstance().App->message
             ( MifMessages::cannotOpenOutputError,
               StringMessageArg( MifDoc::curInstance().App->convertInput( bookFileLoc.data() ) ),
               ErrnoMessageArg( errno ) );
        }
    
    } else
        if( BookComponents.size() == 1 )
            BookComponents[0].FileName = outFileName;
        
    for( size_t i = 0; i < BookComponents.size(); i++ )
        BookComponents[i].commit( outDir );
}

void MifDoc::enterBookComponent() {

    TagStream *defaultTagStream = new TagStream( MifDoc::T_indent( 0 ) );
    enterTagStream( *defaultTagStream );
    BookComponent newBookComponent( &defaultTagStream->content() );
    bookComponents().push_back( newBookComponent );
}

void MifDoc::exitBookComponent() {

    exitTagStream();
}

void MifDoc::outAFrame( T_ID ID, MifOutputByteStream &os ) {

    os << '\n' << MifOutputByteStream::INDENT << "<AFrame " << ID << ">";
}

void MifDoc::switchToTagStream( TagStream &tagStream, bool startWithDefaultPgfFormat ) {

    if( tagStream.TagStreamClass == TagStream::tsTextFlow )
        setCurTextFlow( (MifDoc::TextFlow*)&tagStream );
    else
        if( tagStream.TagStreamClass == TagStream::tsCell )
            setCurCell( (MifDoc::Cell*)&tagStream );

    if( tagStream.TagStreamClass == TagStream::tsPara ) {
        MifDoc::Para::currentlyOpened = true;
    } else
        if( startWithDefaultPgfFormat ) {
            tagStream.PgfTagUsed = false;
            MifDoc::ParagraphFormat defaultPgfFormat;
            defaultPgfFormat.setDSSSLDefaults();
            setCurParagraphFormat( defaultPgfFormat );
        }
    setCurOs( tagStream.content().stream() );
}

void MifDoc::enterTagStream( TagStream &tagStream ) {

    TagStreamStack.resize( TagStreamStack.size() + 1 );
    TagStreamStack.back() = &tagStream;
    switchToTagStream( *TagStreamStack.back() );
}

void MifDoc::exitTagStream() {

    bool startWithDefaultPgfFormat = true;

    assert( TagStreamStack.size() > 0 );
    if( TagStreamStack.back()->TagStreamClass == TagStream::tsPara ) {
        MifDoc::Para::currentlyOpened = false;
        startWithDefaultPgfFormat = false;
    }
    TagStreamStack.resize( TagStreamStack.size() - 1 );
    if( TagStreamStack.size() >= 1 )
        switchToTagStream( *TagStreamStack.back(), startWithDefaultPgfFormat );
}

void MifDoc::outTagEnd() { os() << ">"; }

void MifDoc::outHexChar( unsigned code, MifOutputByteStream *o ) {

    MifOutputByteStream &outS = ( o != NULL ) ? *o : os();
    static const char hex[] = "0123456789abcdef";
    outS << "\\x" << hex[(code >> 4) & 0xf] << hex[code & 0xf] << " ";
}

void MifDoc::outHexChar( unsigned code, String<char> &targetString ) {

    static const char hex[] = "0123456789abcdef";
    targetString += '\\'; targetString += 'x';
    targetString += hex[(code >> 4) & 0xf];
    targetString += hex[code & 0xf];
    targetString += ' ';
}

void MifDoc::outSpecialChar( const String<char> &charName, MifOutputByteStream *o ) {

    MifOutputByteStream &outS = ( o != NULL ) ? *o : os();
    outS << '\n' << MifOutputByteStream::INDENT << "<Char " << charName << ">";
}

void MifDoc::outBreakingPara( MifDoc::T_keyword pgfPlacement ) {

        curFormat().setPgfSpBefore( 0 ),
        curFormat().setPgfSpAfter( 0 ),
        curFormat().setPgfPlacement( pgfPlacement );
        Para::outSimpleProlog( os() );
        curFormat().out( os() );
        ParaLine::outProlog( os() );
        ParaLine::outEpilog( os() );
        Para::outEpilog( os() );
}

void MifDoc::beginParaLine() {

    MifDoc::ParaLine::outProlog( os() );
}

void MifDoc::endParaLine() {

    MifDoc::ParaLine::outEpilog( os() );
}

void MifDoc::outPageNumber() {

    os() << '\n' << MifOutputByteStream::INDENT
         << "<Variable <VariableName `Current Page #'>>";
}

// --------- MifDoc::ElementSet -----------------------------------------------

MifDoc::ElementSet::ElementSet() {}

int MifDoc::ElementSet::bookComponentIndex( unsigned long groveIndex, unsigned long n ) const {

    unsigned short value;
    if( getValue( groveIndex, n, value ) )
        return int( value & BOOK_COMPONENT_INDEX_M() );

    assert( false );
    return 0;
}

void MifDoc::ElementSet::setBookComponentIndex
 ( unsigned long groveIndex, unsigned long n, int i ) {

    unsigned short value = 0;
    getValue( groveIndex, n, value );
    add( groveIndex, n,
         ( value & REFERENCED_F() ) | (unsigned short)( i & BOOK_COMPONENT_INDEX_M() ) );
}

bool MifDoc::ElementSet::hasBeenReferenced( unsigned long groveIndex, unsigned long n ) const {

    unsigned short value = 0;
    getValue( groveIndex, n, value );
    return ( value & REFERENCED_F() ) != 0 ? true : false;
}

void MifDoc::ElementSet::setReferencedFlag( unsigned long groveIndex, unsigned long n ) {

    unsigned short value = 0;
    getValue( groveIndex, n, value );
    add( groveIndex, n, value | REFERENCED_F() );
}

void MifDoc::ElementSet::add
 ( unsigned long groveIndex, unsigned long n, unsigned short value  ) {

    if( groveIndex >= v_.size() )
        v_.resize( groveIndex + 1 );
    Vector<unsigned short> &elems = v_[groveIndex];
    if( n >= elems.size() ) {
        for( size_t k = 1 + (n - elems.size()); k > 0; --k )
            elems.push_back(0);
    }
    elems[n] = value;
}

bool MifDoc::ElementSet::getValue
 ( unsigned long groveIndex, unsigned long n, unsigned short &result ) const {

  if( groveIndex < v_.size() && n < v_[groveIndex].size() ) {
    result = v_[groveIndex][n];
    return true;        
  } else
    return false;
}

#ifdef DSSSL_NAMESPACE
}
#endif
